//<c>   Copyright 1992-1999 by Gerry J. Danen; all rights reserved.

//<t>   Back up and restore floppy diskette using diskette image files

//      Todo:
//      Allow PKZIP/PKUNZIP (or other compression tool) to compress/decompress
//      the images automatically


#include"gd_tools.h"
#include"gd_extrn.h"
#include"gd_files.h"
#include"gd_kb.h"


#define pgVERS_ "V3.12"
#define pgTITL_ "Read & Write Floppy Disk Images"
#define pgCOPR_ "1992"


#define READ_FLOPPY     0x02                    // Int 0x13 read sector
#define WRITE_FLOPPY    0x03                    // Int 0x13 write sector
int     process_mode ;                          // Holds Int 0x13 function #
int     copies = 1 ;                            // # of copies to make
int     floppy_drive_number ;                   // Drive number: 0=A, 1=B, ...
int     SectorSize = 512 ;
int     Sectors=9;                              // Number of sectors per track
int     Cylinders=80;                           // Number of Cyls per side
int     Heads=2;                                // Number of heads
char    Track[38912] ;                          // Buffer for largest track: 2.88M floppy (1024*19*2)
ULONG   disk_size ;                             // Size of disk
FILE    *fhImageFile ;                          // File pointer for disk image
R_REGS  ;                                       // Registers for _int86() calls
S_SREGS ;                                       // Segment regs. for _int86x()
UCHAR   argDriv[132]="", argMode[132]="", argSize[132]="", argFile[132]="",
        sImageFileName[FS_LEN], sWork[81] ;


void    Syntax( char *ErrMess )
{
    int x ;
    static char *pHelp[] =
    {
        " Program Usage ", "",
        "DISKIMAG <mode> <drive> <size> <filename> <copies>", "",
        "1. process mode    : RF to read floppy",
        "                   : WF to write floppy",
        "2. floppy drive    : A:",
        "                   : B:",
        "3. floppy disk size: 360",
        "                   : 720",
        "                   : 1200",
        "                   : 1440",
        "4. imagefile       : if `RF', name of disk image file to create",
        "                   : if `WF', name of disk image file to read",
        "5. copies          : if `WF', how many floppies to create (default=1)",
        "", "Copyright (c) " pgCOPR_ "-" _THE_COPR_YEAR_ " by " _THE_COMPANY_NAME_, "",
        NULPTR
    };

    sc_help_screen( pHelp );

    LOOP( x, 2, 99, 1 )
    {
        if ( pHelp[x] == NULPTR )
            break ;
        sc_msg_at( x+2, 1, "%s", pHelp[x] );
    }

    if ( strlen(ErrMess) )
        ErrMsg( ErrMess );
    gd_exit( BAD_EXIT, "" );
}


int     process_track( int head, int cyl )
{
    static char far *TrkPtr=Track;
    static int iOK, iErrors ;

    iOK = iErrors = 0 ;

    while ( (! iOK) && (iErrors < 5) )
    {
        _AX = process_mode<<8 | Sectors ;
        _CX = cyl         <<8 | 1 ;
        _DX = head        <<8 | floppy_drive_number ;
        _BX = FP_OFF( TrkPtr );
        _ES = FP_SEG( TrkPtr );
        _int86x( DISKETTE_IO, &r, &r, &s );
        if ( _CFLAG )                           // error occurred
        {                                       // reset drive and try again
            _AH = 0 ;
            _DX = floppy_drive_number ;
            _int86( DISKETTE_IO, &r, &r );
            iErrors ++ ;
        }
        else
        {
            iOK ++ ;
        }
    }
    return( iOK );
}


int     process_disk( STRING f, STRING t )
{
    int head, cyl;

    dr_reset_floppy_controller();               // disk reset

    for ( cyl = 0 ; cyl < Cylinders ; cyl++ )
    {
        for ( head = 0 ; head < Heads ; head++ )
        {
            if ( process_mode == WRITE_FLOPPY )
                fread( Track, SectorSize, Sectors, fhImageFile );
            if ( ! process_track( head, cyl ) )
                return( 0 );
            if ( process_mode == READ_FLOPPY )
                fwrite( Track, SectorSize, Sectors, fhImageFile );
        }
        sc_progress( (cyl*100) / Cylinders );
        if ( kb_Hit() )
        {
            BOOL cancel ;
            FlushKeyboard();
            if ( process_mode == READ_FLOPPY )
                cancel = YesNo( "Cancel the copy process (delete output file)", NO, YES );
            else
                cancel = YesNo( "Cancel the copy process (disk may be useless!)", NO, YES );
            ClearBottom();
            if ( cancel )
            {
                fclose( fhImageFile );
                if ( process_mode == READ_FLOPPY )      // if reading floppy
                    fl_delete( sImageFileName );        //  get rid of file
                gd_exit( BAD_EXIT, "Disk copy cancelled" );
            }
        }
    }
    return( 1 );
}


void    exit_routine( void )
{
    CursOn();
}


void    main( int argc, char *argv[] )
{
    struct _stat fi ;

    fs_init( FALSE, SHAREWARE_VERSION, 0, pgVERS_, pgTITL_, pgCOPR_ );
    NewScreen();
    atexit( exit_routine );                     // set up to close orderly

    ST_CPY( argMode, argv[1] );
    ST_CPY( argDriv, argv[2] );
    ST_CPY( argSize, argv[3] );
    ST_CPY( argFile, argv[4] );
    strupr( argDriv );
    strupr( argMode );
    strupr( argSize );
    strupr( argFile );
    ST_CPY( sImageFileName, argFile );
    if ( argc > 5 )
        copies = atoi( argv[5] );

    if ( st_eq( "RF", argMode ) )               // establish Read/Write mode
        process_mode = READ_FLOPPY ;
    else
    if ( st_eq( "WF", argMode ) )
        process_mode = WRITE_FLOPPY ;
    else
        Syntax( "" );

    floppy_drive_number = (int)(argDriv[0] - 'A'); // which drive

    if ( ! dr_is_floppy( (UCHAR) (floppy_drive_number+1) ) )
        Syntax( "Please indicate drive letter for floppy disk" );

    if ( st_eq("360", argSize) )                // diskette format
        Cylinders = 40 ;
    else
    if ( st_eq("720", argSize) )
        ;
    else
    if ( st_eq("1200", argSize) )
        Sectors = 15 ;
    else
    if ( st_eq("1440", argSize) )
        Sectors = 18 ;
    else
    if ( st_eq("2880", argSize) )
        Sectors = 36 ;
    else
        Syntax( "Wrong floppy disk size" );

    if ( process_mode == READ_FLOPPY )          // open/create disk image file
    {
        copies = 1 ;                            // dumb to do it more than once...
        CursOn();
        fhImageFile = fl_create_or_replace_tty( fhImageFile, sImageFileName, "wb", YES, YES );
        CursOff();
    }
    else
    {
        int fh ;
        fh = _open( sImageFileName, _O_RDONLY );
        if ( fh == -1 )
            Syntax( "Could not open image file" );
        _fstat( fh, &fi );
        _close( fh );
        fhImageFile = fopen( sImageFileName, "rb" );
        if ( ! fhImageFile )
            Syntax( "Error opening image file" );
    }

    disk_size = (ULONG)Cylinders *
                (ULONG)Sectors   *
                (ULONG)Heads     *
                (ULONG)SectorSize ;

    if ( process_mode == WRITE_FLOPPY )         // check file size
    {
        if ( (ULONG) fi.st_size != disk_size )
        {
            UCHAR s1[21], s2[21] ;
            Format_UINT32( (ULONG) fi.st_size, s1 );
            Format_UINT32( disk_size, s2 );
            printf( "Image file size expected: %s.  Found: %s.\n", s2, s1 );
            Syntax( "Size mismatch" );
        }
    }

    NewScreen();

    printf( "Floppy size: %s KBytes\n", argSize );
    printf( "%sing %d cylinders (%d sectors)\n",
            (process_mode==READ_FLOPPY ? "Read" : "Writ"), Cylinders, Sectors );

    sprintf( argSize, "Drive %c:", floppy_drive_number + 'A' );
    sprintf( argMode, "File %s", sImageFileName );
    if ( process_mode == WRITE_FLOPPY )
    {
        strcpy( argFile, argSize );
        strcpy( argSize, argMode );
        strcpy( argMode, argFile );
        if ( copies < 2 )
        {
            if ( ! YesNo( "Are you sure you wish to write to the floppy, destroying current data", NO, YES ) )
                gd_exit( BAD_EXIT, "Aborted disk copy" );
            ClearBottom();
        }
    }

    printf( "Copying %s to %s", argSize, argMode );
    if ( copies > 1 )
        printf( " (%d copies)", copies );
    printf( "\n\nPress a key to interrupt the copy process.\n" );

    if ( copies > 1 )                           // making multiple floppies
    {
        int this ;
        BOOL ok ;
        LOOP( this, 1, copies, 1 )
        {
            sprintf( sWork, "Ready to create floppy disk (%d of %d)", this, copies );
            ok = YesNo( sWork, YES, YES );
            ClearBottom();
            if ( ok )
            {
                fclose( fhImageFile );
                fhImageFile = fopen( sImageFileName, "rb" );
                sc_progress_init( "Copying from %s to %s (%d of %d)", argSize, argMode, this, copies );
                if ( ! process_disk(argSize, argMode) ) // copy to floppy
                    Syntax( "Floppy disk error" );
                sn_beep();                      // beep
                sc_progress_exit( 1 );          // clear progress box
            }
        }
    }
    else                                        // single copy
    {
        sc_progress_init( "Copying from %s to %s", argSize, argMode );
        if ( ! process_disk(argSize, argMode) ) // copy to or from floppy
            Syntax( "Floppy disk error" );
        sn_beep();                              // beep
        sc_progress_exit( 3 );                  // clear progress box
    }

    MoveCursor( vid.rows-3, 1 );

    printf( "\nDo NOT use this program to distribute software illegally!\n" );

    gd_exit( GOOD_EXIT, "" );
}
